(defclass xjcl-hash-class (standard-class)
  ())

(defclass xjcl-hash-object (standard-object)
  (xjcl-hash-slots))

(defmethod validate-superclass
           ((class xjcl-hash-class)
            (superclass standard-class))
  t)

(defmethod initialize-instance :around
  ((class xjcl-hash-class) &rest initargs
   &key direct-superclasses)
  (declare (dynamic-extent initargs))
  (if (loop for class in direct-superclasses
            thereis (subtypep class (find-class 'xjcl-hash-object)))
    (call-next-method)
    (apply #'call-next-method
           class
           :direct-superclasses
           (append direct-superclasses
                   (list (find-class 'xjcl-hash-object)))
           initargs)))

(defmethod reinitialize-instance :around
  ((class xjcl-hash-class) &rest initargs
   &key (direct-superclasses '() direct-superclasses-p))
  (declare (dynamic-extent initargs))
  (if (or (not direct-superclasses-p)
          (loop for class in direct-superclasses
                thereis (subtypep class (find-class 'xjcl-hash-object))))
    (call-next-method)
    (apply #'call-next-method
           class
           :direct-superclasses
           (append direct-superclasses
                   (list (find-class 'xjcl-hash-object)))
           initargs)))

(defclass xjcl-hash-direct-slot-definition
          (standard-direct-slot-definition)
  ())

(defmethod direct-slot-definition-class ((class xjcl-hash-class) &rest initargs)
  (declare (ignore initargs))
  (find-class 'xjcl-hash-direct-slot-definition))

(defclass xjcl-hash-effective-slot-definition 
          (standard-effective-slot-definition)
  ())

(defvar *effective-slot-definition-class*)

(defmethod compute-effective-slot-definition
           ((class xjcl-hash-class) (name t) direct-slot-definitions)
  (let ((*effective-slot-definition-class*
         (if (eq (slot-definition-allocation (first direct-slot-definitions))
                 :xjcl-hash)
           (find-class 'xjcl-hash-effective-slot-definition)
           (find-class 'standard-effective-slot-definition))))
    (call-next-method)))

(defmethod effective-slot-definition-class
           ((class xjcl-hash-class) &rest initargs)
  (declare (ignore initargs))
  *effective-slot-definition-class*)

(defmethod shared-initialize :before ((object xjcl-hash-object) slot-names &rest initargs)
  (declare (ignore slot-names initargs))
  (unless (slot-boundp object 'xjcl-hash-slots)
    (setf (slot-value object 'xjcl-hash-slots)
          (make-hash-table :test #'eq))))

(defmethod slot-value-using-class
           ((class xjcl-hash-class)
            object
            (slot xjcl-hash-effective-slot-definition))
  (multiple-value-bind
      (value present-p)
      (gethash (slot-definition-name slot)
               (slot-value object 'xjcl-hash-slots))
    (if present-p value
      (slot-unbound class object (slot-definition-name slot)))))

(defmethod (setf slot-value-using-class)
           (value
            (class xjcl-hash-class)
            object
            (slot xjcl-hash-effective-slot-definition))
  (setf (gethash (slot-definition-name slot)
                 (slot-value object 'xjcl-hash-slots))
        value))

(defmethod slot-boundp-using-class
           ((class xjcl-hash-class)
            object
            (slot xjcl-hash-effective-slot-definition))
  (nth-value 1 (gethash (slot-definition-name slot)
                        (slot-value object 'xjcl-hash-slots))))

(defmethod slot-makunbound-using-class
           ((class xjcl-hash-class)
            object
            (slot xjcl-hash-effective-slot-definition))
  (remhash (slot-definition-name slot)
           (slot-value object 'xjcl-hash-slots)))

(ignore-errors
  (eval '(progn
           (defclass test-object ()
             ((test-slot :allocation :xjcl-hash))
             (:metaclass xjcl-hash-class))
           (finalize-inheritance (find-class 'test-object))))
  (acknowledge :extensible-allocation))
